/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       entry.c
 * PURPOSE:    Registry filesystem entry and unload
 */

#include "regfs.h"

UNICODE_STRING LinkName;

#pragma code_seg(push, INIT)

NTSTATUS NTAPI DriverEntry(PDRIVER_OBJECT DriverObject, PUNICODE_STRING RegistryPath)
{
	PREGFS_DEVICE_EXTENSION DeviceExtension;
	UNICODE_STRING DeviceName;
	NTSTATUS Status;

	DPRINT("regfs FSD 0.1.2\n");

	RegfsDriverObject = DriverObject;

	DriverObject->DriverUnload = &RegfsUnload;
	DriverObject->DriverStartIo = NULL;
	DriverObject->FastIoDispatch = NULL;

	DriverObject->MajorFunction[IRP_MJ_CREATE] =
	DriverObject->MajorFunction[IRP_MJ_CLEANUP] =
	DriverObject->MajorFunction[IRP_MJ_CLOSE] =
	DriverObject->MajorFunction[IRP_MJ_READ] =
	DriverObject->MajorFunction[IRP_MJ_WRITE] =
	DriverObject->MajorFunction[IRP_MJ_QUERY_INFORMATION] =
	DriverObject->MajorFunction[IRP_MJ_SET_INFORMATION] =
	DriverObject->MajorFunction[IRP_MJ_QUERY_VOLUME_INFORMATION] =
	DriverObject->MajorFunction[IRP_MJ_QUERY_EA] =
	DriverObject->MajorFunction[IRP_MJ_SET_EA] =
	DriverObject->MajorFunction[IRP_MJ_FLUSH_BUFFERS] =
	DriverObject->MajorFunction[IRP_MJ_DIRECTORY_CONTROL] =
	DriverObject->MajorFunction[IRP_MJ_FILE_SYSTEM_CONTROL] =
	DriverObject->MajorFunction[IRP_MJ_QUERY_SECURITY] =
	DriverObject->MajorFunction[IRP_MJ_SET_SECURITY] =
		RegfsInitDispatch(DmObjKernelHandle, &Status);

	if(!NT_SUCCESS(Status))
	{
		 DPRINT("Error initializing dispatch. (Status %x)\n", Status);
		 return Status;
	}

	DriverObject->DriverUnload = NULL;

	Status = RegfsInitFcbTable();

	if(!NT_SUCCESS(Status))
	{
		RegfsCleanupWorkThreads();
		return Status;
	}

	Status = ObOpenObjectByPointer(PsGetCurrentProcess(), DefaultObjectAttributes, NULL, PROCESS_DUP_HANDLE, NULL, KernelMode, &SystemProcessHandle);

	if (!NT_SUCCESS(Status))
	{
	   DPRINT("Failed to open system process! (Status %x)\n", Status);
	   RegfsCleanupFcbTable();
	   RegfsCleanupWorkThreads();
	   return Status;
	}

	RtlInitUnicodeString(&DeviceName, L"\\Device\\Regfs");
	//Status = IoCreateDevice(DriverObject, sizeof(REGFS_DEVICE_EXTENSION), &DeviceName, FILE_DEVICE_DISK, 0, FALSE, &RegfsDevice);
	Status = IoCreateDevice(DriverObject, sizeof(REGFS_DEVICE_EXTENSION), &DeviceName, FILE_DEVICE_FILE_SYSTEM, 0, FALSE, &RegfsDeviceObject);
	if (!NT_SUCCESS(Status))
	{
	   DPRINT("Failed to create regfs device! (Status %x)\n", Status);
	   RegfsCleanupFcbTable();
	   RegfsCleanupWorkThreads();
	   ZwClose(SystemProcessHandle);
	   return Status;
	}

	DeviceExtension = RegfsDeviceObject->DeviceExtension;
	/* initialize the device object */
	//RegfsDeviceObject->Flags = DO_DIRECT_IO | DO_POWER_PAGABLE;
	//RegfsDeviceObject->Flags = DO_BUFFERED_IO | DO_POWER_PAGABLE;
	RegfsDeviceObject->Flags |= DO_POWER_PAGABLE; //NETHIER_IO
	RegfsDeviceObject->StackSize = 1; //just regfs
	RegfsDeviceObject->AlignmentRequirement = FILE_BYTE_ALIGNMENT;

	{
		UNICODE_STRING LinkDestName;
		//RtlInitUnicodeString(&LinkDestName, L"\\Device\\Regfs\\REGISTRY\\MACHINE\\SOFTWARE\\Microsoft");
		//RtlInitUnicodeString(&LinkDestName, L"\\Device\\Regfs\\REGISTRY\\MACHINE\\SOFTWARE");
		RtlInitUnicodeString(&LinkDestName, L"\\Device\\Regfs\\REGISTRY");
		RtlInitUnicodeString(&LinkName, L"\\DosDevices\\R:");
		IoCreateSymbolicLink(&LinkName, &LinkDestName);
	}

	DriverObject->DriverUnload = &RegfsUnload;

	return STATUS_SUCCESS;
}

#pragma code_seg(pop, INIT)

VOID NTAPI RegfsUnload(IN PDRIVER_OBJECT DriverObject)
{
	IoDeleteSymbolicLink(&LinkName);
	IoDeleteDevice(RegfsDeviceObject);
	RegfsCleanupWorkThreads();
	RegfsCleanupExWorkItems();
	RegfsCleanupFcbTable();
	ZwClose(SystemProcessHandle);

	DPRINT("regfs FSD unloaded\n");
}

NTSTATUS NTAPI RegfsNotifyDirectory(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	//return RegfsCompleteStatus(Irp, FALSE, STATUS_NOT_IMPLEMENTED, "Error: Directory notification not implemented.");
	return RegfsCompleteStatus(Irp, FALSE, STATUS_NOT_IMPLEMENTED, "NoNotify");
}

//dispatch to queryinf.c or notify.c
NTSTATUS NTAPI RegfsDirectoryControl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	PIO_STACK_LOCATION IrpSp;
	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);

	switch(IrpSp->MinorFunction)
	{
	case IRP_MN_NOTIFY_CHANGE_DIRECTORY:
		return RegfsNotifyDirectory(DeviceObject, Irp, IoBuffer, IoBufferLength, SecurityContext);
	case IRP_MN_QUERY_DIRECTORY:
		return RegfsQueryDirectory(DeviceObject, Irp, IoBuffer, IoBufferLength, SecurityContext);
	default:
		return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "Error: Unspported directory control function.");
	}
}