/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       fsctl.c
 * PURPOSE:    Registry filesystem special control
 */

#include "regfs.h"

ULONG const FSCTL_REGFS_TEST1 = CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 0x800, METHOD_NEITHER, FILE_ANY_ACCESS);

NTSTATUS NTAPI RegfsTest1(PIRP Irp)
{
	HANDLE H1, H2;
	IO_STATUS_BLOCK Iosb;
	UNICODE_STRING Fn1, Fn2;
	OBJECT_ATTRIBUTES Oa1, Oa2;
	CHAR Buff1[1024], Buff2[1024];
	PFILE_ALL_INFORMATION Foa1 = (PFILE_ALL_INFORMATION)Buff1, Foa2 = (PFILE_ALL_INFORMATION)Buff2;
	NTSTATUS Status;

	//RtlInitUnicodeString(&Fn1, L"\\Device\\Regfs\\REGISTRY\\MACHINE\\SOFTWARE\\Test\\Value");
	//RtlInitUnicodeString(&Fn2, L"\\Device\\HarddiskVolume1\\SOFTWARE\\Test\\Value");

	//RtlInitUnicodeString(&Fn1, L"\\Device\\Regfs\\REGISTRY\\MACHINE\\SOFTWARE\\Test\\");
	//RtlInitUnicodeString(&Fn2, L"\\Device\\HarddiskVolume1\\SOFTWARE\\Test\\");

	RtlInitUnicodeString(&Fn1, L"\\??\\R:\\Test\\");
	RtlInitUnicodeString(&Fn2, L"\\??\\N:\\Test\\");

	InitializeObjectAttributes(&Oa1, &Fn1, OBJ_CASE_INSENSITIVE, NULL, NULL);
	InitializeObjectAttributes(&Oa2, &Fn2, OBJ_CASE_INSENSITIVE, NULL, NULL);

	DbgBreakPoint();

	Status = ZwOpenFile(&H1, FILE_ALL_ACCESS, &Oa1, &Iosb, 0, FILE_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_NONALERT);
	Status = ZwOpenFile(&H2, FILE_ALL_ACCESS, &Oa2, &Iosb, 0, FILE_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_NONALERT);

	Status = ZwQueryInformationFile(H1, &Iosb, Foa1, 1024, FileAllInformation);
	Status = ZwQueryInformationFile(H2, &Iosb, Foa2, 1024, FileAllInformation);

	return RegfsCompleteStatus(Irp, TRUE, Status, NULL);
}

NTSTATUS NTAPI RegfsTest2(PIRP Irp)
{
	HANDLE H1, H2;
	IO_STATUS_BLOCK Iosb;
	UNICODE_STRING Fn1, Fn2;
	OBJECT_ATTRIBUTES Oa1, Oa2;
	CHAR Buff1[1024], Buff2[1024];
	PFILE_FULL_DIR_INFORMATION Foa1 = (PFILE_FULL_DIR_INFORMATION)Buff1, Foa2 = (PFILE_FULL_DIR_INFORMATION)Buff2;
	NTSTATUS Status;

	RtlInitUnicodeString(&Fn1, L"\\Device\\Regfs\\REGISTRY\\MACHINE\\SOFTWARE\\Test");
	RtlInitUnicodeString(&Fn2, L"\\Device\\HarddiskVolume1\\SOFTWARE\\Test");

	InitializeObjectAttributes(&Oa1, &Fn1, OBJ_CASE_INSENSITIVE, NULL, NULL);
	InitializeObjectAttributes(&Oa2, &Fn2, OBJ_CASE_INSENSITIVE, NULL, NULL);

	DbgBreakPoint();

	Status = ZwOpenFile(&H1, FILE_ALL_ACCESS, &Oa1, &Iosb, 0, FILE_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_NONALERT);
	Status = ZwOpenFile(&H2, FILE_ALL_ACCESS, &Oa2, &Iosb, 0, FILE_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_NONALERT);

	Status = ZwQueryDirectoryFile(H1, NULL, NULL, NULL, &Iosb, Foa1, 1024, FileFullDirectoryInformation, FALSE, NULL, TRUE);
	Status = ZwQueryDirectoryFile(H2, NULL, NULL, NULL, &Iosb, Foa2, 1024, FileFullDirectoryInformation, FALSE, NULL, TRUE);

	return RegfsCompleteStatus(Irp, TRUE, Status, NULL);
}

NTSTATUS NTAPI RegfsFsControl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN OUT PVOID IoBuffer, IN ULONG IoBufferLength, IN PSECURITY_SUBJECT_CONTEXT SecurityContext OPTIONAL)
{
	NTSTATUS Status;
	PIO_STACK_LOCATION IrpSp;
	PFILE_OBJECT FileObject;

	IrpSp = (PIO_STACK_LOCATION)IoGetCurrentIrpStackLocation(Irp);
	FileObject = IrpSp->FileObject;

	if(HDRHELP_EXT_IRPSP(IrpSp)->Parameters.FileSystemControl.FsControlCode == FSCTL_REGFS_TEST1)
		return RegfsTest1(Irp);
	
	DPRINT("RegfsFsControl DeviceObject:%p, Irp:%p FileName:%wZ\n", DeviceObject, Irp, &FileObject->FileName);
	return RegfsCompleteStatus(Irp, FALSE, STATUS_INVALID_DEVICE_REQUEST, "Invalid request.");
}