/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       inline.h
 * PURPOSE:    Registry filesystem lightweight inlined support functions
 */

#define AlignUpTo(_Val, _Shift) (((_Val) + ((1 << (_Shift)) - 1)) & ~((1 << (_Shift)) - 1))

//returns amount adjusted by, if any
SIZE_T FORCEINLINE NTAPI AlignUpToAndZero(IN OUT PVOID *Address, IN PVOID MaxAddress, IN UCHAR Shift)
{
	SIZE_T Diff = AlignUpTo((ULONG_PTR)*Address, Shift) - (ULONG_PTR)*Address;

	Diff = min(Diff, (ULONG_PTR)MaxAddress - (ULONG_PTR)Address); //avoid overflow

	if(Diff)
		RtlZeroMemory(*Address, Diff);

	(ULONG_PTR)*Address += Diff;

	return Diff;
}

HANDLE FORCEINLINE NTAPI RegfsGetDfcbKeyHandle(IN PREGFS_DFCB Dfcb)
{
	return Dfcb->Flags & REGFS_DFCB_DIRECTORY ? Dfcb->Dcb.KeyHandle : Dfcb->Fcb.Directory->Dcb.KeyHandle;
}

NTSTATUS FORCEINLINE NTAPI RegfsCompleteStatus(IRP *Irp, BOOLEAN DidIo, NTSTATUS Status, PCHAR Message)
{
#if DBG
	//the message is only used in debug builds
	Irp->Tail.Overlay.DriverContext[0] = Message;
#endif
	Irp->Tail.Overlay.DriverContext[1] = DidIo ? NULL : (PVOID)1;
	return Status;
}

NTSTATUS FORCEINLINE NTAPI RegfsCompleteKeyStatus(IRP *Irp, BOOLEAN DidIo, NTSTATUS Status, PCHAR Message)
{
	return RegfsKeyStatusToFileStatus(RegfsCompleteStatus(Irp, DidIo, Status, Message));
}

BOOLEAN FORCEINLINE NTAPI RangeInRange(ULONG_PTR ValMin, ULONG_PTR ValMax, ULONG_PTR Min, ULONG_PTR Max)
{
	return ValMin <= ValMax && ValMin >= Min && ValMax <= Max;
}

BOOLEAN FORCEINLINE NTAPI IsBufferSizeStatus(NTSTATUS Status)
{
	return Status == STATUS_BUFFER_TOO_SMALL || Status == STATUS_BUFFER_OVERFLOW;
}

//it's like TRY_LEAVE
#define EXIT_AND(expr) {expr; goto exit;}