/*
 * COPYRIGHT:  2007 Christopher Smith
 * LICENCE:    GPL v2: See COPYING in the top level directory
 * PROJECT:    regfs
 * FILE:       osver.h
 * PURPOSE:    Registry filesystem OS version specific preprocessing and definitions
 */

//KERNEL_VERSION is a macro whose value is 0xAABB where A is the major revision and B is the minor revision
//i.e. NT 3.51 is 0x0351, NT4 is 0x0400, XP is 0x0501

#ifndef KERNEL_VERSION

	//set it to the oldest supported to be safe
	#define KERNEL_VERSION 0x351

#endif

#if KERNEL_VERSION < 0x501

	#define ExFreePoolWithTag(_Buffer, _Tag) ExFreePool(_Buffer)

#endif //KERNEL_VER < 0x501


#if KERNEL_VERSION >= 0x500

	#define MmMapLockedPagesSafest(_Mdl, _AccessMode) MmMapLockedPagesSpecifyCache(_Mdl, _AccessMode, MmCached, NULL, FALSE, NormalPagePriority)

#else

	#define MmMapLockedPagesSafest(_Mdl, _AccessMode) ((_Mdl)->MdlFlags |= MDL_MAPPING_CAN_FAIL, MmMapLockedPages((_Mdl), (_AccessMode)))

#endif

#if KERNEL_VERSION >= 0x502

	typedef KGUARDED_MUTEX SIMPLE_MUTEX;
	typedef PKGUARDED_MUTEX PSIMPLE_MUTEX;
	#define AcquireSimpleMutex(_Mutex) KeAcquireGuardedMutex(_Mutex)
	#define TryAcquireSimpleMutex(_Mutex) KeTryToAcquireGuardedMutex(_Mutex)
	#define InitializeSimpleMutex(_Mutex) KeInitializeGuardedMutex(_Mutex)
	#define ReleaseSimpleMutex(_Mutex) KeReleaseGuardedMutex(_Mutex)
	#define SIMPLE_MUTEX_STORAGE NonPagedPool

#elif KERNEL_VERSION == 0x501

	typedef FAST_MUTEX SIMPLE_MUTEX;
	typedef PFAST_MUTEX PSIMPLE_MUTEX;
	#define AcquireSimpleMutex(_Mutex) ExAcquireFastMutex(_Mutex)
	#define TryAcquireSimpleMutex(_Mutex) ExTryToAcquireFastMutex(_Mutex)
	#define InitializeSimpleMutex(_Mutex) ExInitializeFastMutex(_Mutex)
	#define ReleaseSimpleMutex(_Mutex) ExReleaseFastMutex(_Mutex)
	#define SIMPLE_MUTEX_STORAGE NonPagedPool

#else //KERNEL_VERSION < 0x501

	LONGLONG SELECTANY CONST SIMPLE_MUTEX_ZERO_WAIT = 0;

	typedef KMUTEX SIMPLE_MUTEX;
	typedef PRKMUTEX PSIMPLE_MUTEX;
	#define AcquireSimpleMutex(_Mutex) KeWaitForSingleObject(_Mutex, Executive, KernelMode, FALSE, NULL)
	#define TryAcquireSimpleMutex(_Mutex) (KeWaitForSingleObject(_Mutex, Executive, KernelMode, FALSE, (PLARGE_INTEGER)&SIMPLE_MUTEX_ZERO_WAIT) == STATUS_SUCCES)
	#define InitializeSimpleMutex(_Mutex) KeInitializeMutex(_Mutex, 0)
	#define ReleaseSimpleMutex(_Mutex) KeReleaseMutex(_Mutex, FALSE)
	#define SIMPLE_MUTEX_STORAGE NonPagedPool

#endif

PSIMPLE_MUTEX FORCEINLINE NTAPI AllocSimpleMutex(ULONG Tag)
{
	PSIMPLE_MUTEX Mutex = ExAllocatePoolWithTag(SIMPLE_MUTEX_STORAGE, sizeof(SIMPLE_MUTEX), Tag);
	if(Mutex)
		InitializeSimpleMutex(Mutex);
	return Mutex;
}

VOID FORCEINLINE NTAPI FreeSimpleMutex(PSIMPLE_MUTEX Mutex, ULONG Tag)
{
	ExFreePoolWithTag(Mutex, Tag);
}